; comments prefixed by  were added by JPM.
; modif 12F675 -> 16F627
; Ver 8.0: mixte 627/675/629 et rglage des position avec action.
; Date:	April, 3, 2005
;  December 2007
; This is a simple single switch DCC decoder with the 16F627
; It uses the internal osc. Execution cycle is 1us.
; No CVs to fiddle with, programmable directly from switch control device:
; decoder address and switch impulse duration (with constant and re-trigerrable output options)

; It incorporates ideas from D.Probst DCC decoder project

; Copyright (C) 2005 Alex Lagutin <alex@rogueamoeba.com>
; Parts copyright (C) 2003  Heiko Schroeter
; Parts copyright (C) Dean Probst

;
; How to program this decoder (using Roco RouteControl Keyboard as an example)
;
; Accessory address:
; Press and hold ADDRESS button on a decoder
; type in desired number and press OK or ESC on a keyboard
; Release addr_btn
; Possible address ranges: 1-255, default address - 1
; Lenz addresses are shifted by 3
; eg: "5" in source code -> "4" in EEPROM -> "1" on LH100	
; Switch impulse duration:
; Press and hold TIME button on a decoder
; type in desired number and press OK or ESC on a keyboard
; Release time_btn
; Possible duration ranges: 1-255, default duration - 1 (20ms)
; duration calculated by decoder as Value*20ms
; 20ms = 22us (base cycle) * 255 (timer low) * 2 (scaler bit) a ne fait pas le compte!
; Duration value of 255 has special meaning: Constant Output
; i.e. output pin remains on until switched to another position
; Attention: here LH100 must address switch #251/252 !
; if duration value is even number, then output becomes re-trigerrable
; i.e. same command repeated will trigger output again and again
;-----------------------------------------------
; About DCC:according to NMRA RP-921,
; the structure of the DCC packet is:
; packet = preamble, 0, data1, 0, data2, 0, (more data, 0), error detection byte, 1
;-----------------------------------------------
; with data1 = 0: broadcast
;-----------------------------------------------
; with data1 = 1 to 127: multi-function decoder with 7 bits address:
; 1111111111 0 0AAAAAAA 0 01DUSSSS 0 EEEEEEEE 1
;-----------------------------------------------
; with data1 = 128 to 191:
; Basic Accessory Decoder Format with 9 bits address (this decoder):
; 1111111111 0 10AAAAAA 0 1AAACDDD 0 EEEEEEEE 1
; C = on/off, DDD = outlet (lsb = pair)
; According to NMRA RP-921 the structure of the
; Basic Accessory Decoder Packet Format is:
; Preamble (0) ByteOne (0) ByteTwo (0) ByteThree (1)
; with:
; ByteOne = 1 0 A5 A4 A3 A2 A1 A0
; ByteTwo = 1 A8 A7 A6 C D2 D1 D0
; ByteThree = error detection byte
; A8-A0 = decoder address (with bits A8-A6 inverted)
; D2-D0 = decoder output to be activated/deactivated
; C = activate/deactivate
; or:
; D2-D1 = switch to be operated
; D0 = switch coil to be energized
;
; In the present "single switch" decoder this is interpreted as:
; A5 A4 A3 A2 A1 A0 D2 D1 = switch address (0-255)
; D0 = switch coil
; C is ignored (always 1 in Lenz)
; 
;---
; still with data1 = 128 to 191:
; Extended Accessory Decoder Format:
; 1111111111 0 10AAAAAA 0 0AAA0AA1 0 000XXXXX 0 EEEEEEEE 1
;-----------------------------------------------
; with data1 = 192 to 231: multi-function decoder with14 bits address
; 1111111111 0 110AAAAA etc
;-----------------------------------------------
; data1 = 232 to 254: reserved
;-----------------------------------------------
; data1 = 255: idle packet 
;-----------------------------------------------
	variable mypic = 675	;16F627 or 12F675 or 12F629
; PIC is 16F627
	#if mypic == 627
		LIST		P=PIC16F627
		#INCLUDE	<P16F627.INC>
  		__CONFIG _BODEN_ON & _CP_OFF & _PWRTE_ON & _WDT_OFF & _MCLRE_ON & _DATA_CP_OFF & _LVP_OFF & _XT_OSC ;16F627
	#endif
; PIC is 12F675
	#if mypic == 675
		LIST		P=PIC12F675
		#INCLUDE	<P12F675.INC>
		__CONFIG _BODEN_ON & _CP_OFF & _PWRTE_ON & _WDT_OFF & _MCLRE_OFF & _INTRC_OSC_NOCLKOUT ; 12F675 MCLRE modifi
	#endif
; PIC is 12F629
	#if mypic == 629
		LIST		P=PIC12F629
		#INCLUDE	<P12F629.INC>
		__CONFIG        _BODEN_ON & _CP_OFF & _PWRTE_ON & _WDT_OFF & _MCLRE_OFF & _INTRC_OSC_NOCLKOUT ; 12F629 MCLRE modifi
	#endif
;==========================================================================
; E/S du K8048 Velleman pour m.a.p.
	#if mypic == 627
		#define LED1		PORTB,0	; RB0 - Velleman pour m.a.p.
		#define LED2		PORTB,1	; RB1 - Velleman pour m.a.p.
		#define LED3		PORTB,2	; RB2 - Velleman pour m.a.p.
		#define LED4		PORTB,3	; RB3 - Velleman pour m.a.p.
		#define	LED5		PORTB,4	; RB4 - Velleman pour m.a.p.
		#define	LED6		PORTB,5	; RB5 - Velleman pour m.a.p.
		#define SW1			PORTA,0	; RA0 - SW 1		
		#define SW2			PORTA,1	; RA1 - SW 2
		#define SW3			PORTA,2	; RA2 - SW 3
		#define SW4			PORTA,3	; RA3 - SW 4
	#endif
; E/S du dcodeur sur K8048
	#if mypic == 627
		#define DCCin		SW1		; entre du signal DCC		
		#define AuxButton	SW2		; bouton auxiliaire
		#define AddressButton	SW3	; rglage de l'adresse
		#define TimeButton	SW4		; rglage de la dure
		#define ServoOut	LED1	; sortie servo
		#define	LightOut	LED2	; fanaux
		#define	Output_A	LED5	; sortie A pour m.a.p.
		#define	Output_B	LED6	; sortie B pour m.a.p.
		#define RAMtop		0x7F
; E/S du dcodeur sur 12F675 ou 12F629	
	#else
		#define DCCin		GPIO,2	; entre du signal DCC		
		#define AuxButton	GPIO,3	; bouton auxiliaire - pas de MCLR mais une pull-up externe
		#define AddressButton	GPIO,4	; rglage de l'adresse
		#define TimeButton	GPIO,5	; rglage de la dure
		#define ServoOut	GPIO,1	; sortie servo
		#define	LightOut	GPIO,0	; fanaux
		#define RAMtop		0x5F
	#endif

; dclarations communes
#define	acc_addr_ee	0		; location to store ACC_ADDR in EEPROM
#define	acc_time_ee	1		; location to store ACC_ON_TIME in EEPROM
#define ServoPW_A_ee 2		; location to store ServoPW_A in EEPROM
#define ServoPW_B_ee 3		; location to store ServoPW_B in EEPROM
#define	AddrBtnPreFlag	BTN_STATE_PRE,2
#define	AddrBtnPostFlag	BTN_STATE_POST,2		
#define	TimeBtnPreFlag	BTN_STATE_PRE,3
#define	TimeBtnPostFlag	BTN_STATE_POST,3
#define nopreamb	D'19'		; No. of preamble bits = 20
; Bit Positions in TEST_BITS
#define bitrec		0		; What value has just rec. bit
#define byte4		1		; 4 or 5 byte packet received

	cblock	0x20
ACC_ADDR	; our accessory address
ACC_ON_TIME	; time to keep output pin up
ACTIVE_PAIR	; last pair active. Bit 0 = coil, bit 7 = see do_switch 
ON_TIMER_CTRL	; on-time counter control. Bit 0 = scaler bit = divider by 2, bit 7 = timer enabled 
ON_TIMER_HI	; on-time counter hi-part
ON_TIMER_LO	; on-time counter low-part. Always counts 255
WR_EEDATA	; register used to pass data to write into EEPROM
TEST_BITS	; Contains various bits to test
ENDVAL		; Offset to byte check routines
PRECOUNT	; Preamble counter
SAMPLES		; How many samples taken for one or zero
STATE		; Where are we in the DCC package
DATA1		; First transm. byte
DATA2		; Second trans. byte
DATA3		; Third transm. byte
DATA4		; Fourth transm. byte
DATA5		; Scratch Register
BTN_STATE_PRE	; State of buttons at the beginning of decoding
BTN_STATE_POST	; State of buttons at the end of decoding
TEMP
TEMP1
ServoPW_A	; dure d'activation en position A
ServoPW_B	; dure d'activation en position B
ServoPWM	; compteur de dure d'activation. Dcompte de ServoPositionA/B  0.
ZeroCtr		; talonnage du point milieu
Scaler		; multiplicateur de la dure d'activation
	endc

	org		0x2100	; EEPROM
	de		4	;0xFF	; address (FF = blank eeprom, causes initial setting) 4->1/LH100
	de		d'252'	; time
	de		d'32'	; position A
	de		d'224'	; position B
; macro to write W in eeprom at location: location
store_W_in_eeprom_at macro location
	movwf	WR_EEDATA
	movlw	location
	call	write_eeprom
	endm

; Macro for generating the Output switching timing
switch	macro	jumpto
	local	exit11, exit13, exit15
; tout est supprim, et un bouton/cavalier de mise au centre ajout
	#if mypic == 627	; boutons au +
		btfss	AuxButton			;8/9	
		goto	exit11				;9,10	pas enfonc -> fin
		btfsc	AddressButton		;10/11	enfonc: et AB?
		goto	exit13				;11,12	enfonc: trait ailleurs
		btfsc	TimeButton			;12/13	et TB?
		goto	exit15				;13,14	enfonc: trait ailleurs
	#else				; boutons au -
		btfsc	AuxButton			;8/9
		goto	exit11				;9,10	pas enfonc -> fin
		btfss	AddressButton		;10/11	enfonc: et AB?
		goto	exit13				;11,12	enfonc: trait ailleurs
		btfss	TimeButton			;12/13	et TB?
		goto	exit15				;13,14	enfonc: trait ailleurs
	#endif
;MidPoint:
	movlw	D'128'
	movwf	ServoPWM		; load pulse duration
	clrf	ON_TIMER_LO		; clear timer lo 
;	movf	ACC_ON_TIME,W
	movlw	d'252'			; 
	movwf	ON_TIMER_HI		; load timer hi 
	call	set_up_timer
; attente relachement du bouton
	#if mypic == 627	; bouton au +
		btfsc	AuxButton		 
		goto	$-1
	#else				; bouton au -
		btfss	AuxButton
		goto	$-1
	#endif

	goto	jumpto

exit11:
	goto	$+1					;11,12
exit13:
	goto	$+1					;13,14
exit15:
	goto	$+1					;15,16
	goto	$+1					;17,18
	nop							;19
	goto	jumpto				;20,21

	endm

;************* Here we start ! ******************************************************************************

	org	0
; for 12F675 add:
	#if mypic == 675
		bsf		STATUS,RP0	; Bank 1
		call	0x3ff		; Get calibration value
		movwf	OSCCAL		; Set calibration
		clrf	ANSEL		; disable
		clrf	VRCON		; the ADC
		bcf		STATUS,RP0	; Bank 0
	#endif
; for 12F629 add:
	#if mypic == 629
		bsf		STATUS,RP0	; Bank 1
		call	0x3ff		; Get calibration value
		movwf	OSCCAL		; Set calibration
		bcf		STATUS,RP0	; Bank 0
	#endif
; for 16F627, nothing specific
	goto	start		

;************************************************************************************************************
;Value checks for correct bit and jumps to where we are in DCC package
;Computed goto has to reside in lower half of page, that's why it is here

Value:
	clrf	SAMPLES		; 6
	movf	STATE,W		; 7 W = STATE
	addwf	PCL,F		; 8 PCL = PCL + W
; Jump table to routines from where we do a RETURN !
; cycl. to here ->  cycl for rout.
	goto	waitn		; 9,10
	goto	waitlo
	goto	testlo
; First byte
	goto	bitset
	goto	lastv
	goto	bitset
	goto	lastv
	goto	bitset
	goto	lastv
	goto	bitset
	goto	lastv
	goto	bitset
	goto	lastv
	goto	bitset
	goto	lastv
	goto	bitset
	goto	lastv
	goto	bitset
	goto	lastx		; Byte received
	goto	end11		; Check first half bit of byte
	goto	end12		; Check second half bit of byte
	goto	end21
	goto	end22
	goto	end31
	goto	end32
	goto	end41
	goto	end42
	goto	end51
	goto	end52

;************************************************************************************
start:
; disable comparator (all three pics)
	movlw	0x7			; turn digital i/o
	movwf	CMCON		; on GP0/1/2
; initialisations en bank 1
;	bsf		STATUS, RP0		; Bank 1
	;bsf	OPTION_REG, NOT_RBPU	; disable all pull-ups 
	; disable everything = power-on status
	;clrf	INTCON		; disable interrupts
	; disable all interrupts = power-on status
	#if mypic == 627
		; initialisations en bank 1
		bsf		STATUS, RP0		; Bank 1
		movlw	b'11111111'	;all RA ports are inputs 
		movwf	TRISA
		movlw	b'00000000'	;all RB ports are outputs 
		movwf	TRISB
		; initialisations en bank 0
		bcf		STATUS,RP0		; Bank 0
		clrf 	PORTB
	#else
		clrf	GPIO		
		; initialisations en bank 1
		bsf		STATUS,RP0		; Bank 1
		bcf		OPTION_REG,NOT_GPPU	; enable internal pull-ups
		movlw	b'11111100'		; GP0 and GP1 out, other in
		movwf	TRISIO		
		movlw	b'00110000'		; pull-ups on GP 4,5. External on GP3.
		movwf	WPU
		; initialisations en bank 0
		bcf		STATUS,RP0		; Bank 0
	#endif
; initialisations en bank 0
;	bcf		STATUS,RP0		; Bank 0
;	clrf 	PORTB
; clear RAM
	movlw	0x20			; bottom of RAM
	movwf	FSR
Next1:
	clrf	INDF
	incf	FSR,F
	movlw	RAMtop			; end of RAM
	subwf	FSR,W
	btfss	STATUS,C
	goto	Next1
	clrf	INDF

	movlw	nopreamb		; Preamble = 20 half bits
	movwf	PRECOUNT

	call	check_power_up

;************************************************************************************************************
; High Bit Level Section
starthi:
	call	Value			; 4,5 pass here when the signal has changed from lo to hi 
conthi:
	btfss	DCCin			; 1	HighLevel ?
	goto	startlo			; 2,3	No
	incf	SAMPLES, F		; 3	Increment SAMPLES Counter
	bcf	TEST_BITS, bitrec	; 4	clear BITREC
	movlw	0xFD			; 5	two or more samples are a ZERO
	addwf	SAMPLES, W		; 6	add the received pulses
	btfss	STATUS, C		; 7	
	bsf	TEST_BITS, bitrec	; 8	less then two SAMPLES is a one
	switch	conthi			; include macro terminated by goto conthi: loop while signal is hi 
;----------------------------------------------------------------------------------------------------
; Low Level Bit Section
startlo:
	call	Value			; 4,5
contlo:
	btfsc	DCCin			; 0
	goto	starthi			; 1,2
	incf	SAMPLES,F		; 2
	bcf	TEST_BITS,bitrec	; 3
	movlw	0xFD			; 4
	addwf	SAMPLES,W		; 5
	btfss	STATUS,C		; 6
	bsf	TEST_BITS,bitrec	; 7
	switch	contlo
;----------------------------------------------------------------------------------------------------
; Frame error in packet
frameerr:
	movlw	nopreamb
	movwf	PRECOUNT
	clrf	STATE
	clrf	ENDVAL
	clrf	DATA1
	clrf	DATA2
	clrf	DATA3
	clrf	DATA4
	clrf	DATA5
	clrf	SAMPLES
	movlw	0xF0			; cut of low nibble in TEST_BITS Byte
	andwf	TEST_BITS,F
	return
;************************************************************************************************************
; waitn waits for 20 half bits of the preamble

waitn:
	goto	$+1			;11,12
	btfss	TEST_BITS,bitrec	;13   there are only ones in preamble
	goto	waitn1			;14,15
	decfsz	PRECOUNT,F		;16
	goto	ret4			;16,17
	movlw	nopreamb		;17
	movwf	PRECOUNT		;18
	incf	STATE,F			;19
	retlw	0			;20,21
ret4:
	goto	$+1			;18,19
	retlw	0			;20,21

waitn1:
	movlw	nopreamb		;16
	movwf	PRECOUNT		;17
	goto	$+1			;18,19
	retlw	0			;20,21

;************************************************************************************************************
; waitlo waits for low half bit after preamble

waitlo:
	goto	$+1			;11,12
	nop				;13

waitlo1:
	goto	$+1			;14,15
	nop				;16
	btfsc	TEST_BITS,bitrec	;17	must be a zero
	goto	ret5			;18	might be long preamble
	incf	STATE,F			;19
ret5:
	retlw	0			;20,21

;************************************************************************************************************
; testlo test second half bit of start bit
testlo:
	goto	$+1			;11,12
	goto	$+1			;13,14
	goto	$+1			;15,16
	btfsc	TEST_BITS,bitrec	;17	must be a zero
	goto	frameerr		;18
	incf	STATE,F			;19
	retlw	0			;20,21

;************************************************************************************************************
; bitset takes first half bit of a bit in the byte
bitset:	
	incf	STATE,F			;11
	bcf	STATUS,C		;12
	btfsc	TEST_BITS,bitrec	;13
	bsf	STATUS,C		;14
	rlf	DATA5,F			;15
	goto	$+1			;16,17
	goto	$+1			;18,19
	retlw	0			;20,21

;************************************************************************************************************
; lastv checks that first half bit = second half bit
lastv:
	goto	$+1			;11,12
	nop				;13
	incf	STATE,F			;14
	btfss	TEST_BITS,bitrec	;15
	goto	lastv1			;16,17
	btfss	DATA5,0			;17
	goto	frameerr		;18,19	error if not equal
	nop				;19
	retlw	0			;20,21

lastv1:
	btfsc	DATA5,0			;18
	goto	frameerr		;19	error if not equal
	retlw	0			;20,21
;************************************************************************************************************
; lastx checks that first half bit = second half bit
; and sets offset to byte check routine
lastx:	
	incf	ENDVAL,F		;11
	movf	ENDVAL,W		;12
	addwf	STATE,F			;13
	nop				;14
	btfss	TEST_BITS,bitrec;15
	goto	lastx1			;16,17
	btfss	DATA5,0			;17
	goto	frameerr		;18	error if not equal
	nop				;19
	retlw	0			;20,21

lastx1:
	btfsc	DATA5,0			;18
	goto	frameerr		;19	error if not equal
	retlw	0			;20,21

;************************************************************************************************************
; end11 end of first byte there must be zero
; end11 first half bit. end12 second half bit
end11:
	btfsc	TEST_BITS,bitrec;11
	goto	frameerr		;12,13
	incf	STATE,F			;13
	incf	ENDVAL,F		;14
	goto	$+1			;15,16
	goto	$+1			;17,18
	nop				;19
	retlw	0			;20,21

end12:	
	btfsc	TEST_BITS,bitrec	;11
	goto	frameerr		;12
	movf	DATA5,W			;13
	movwf	DATA1			;14
	movlw	0x03			;15	point STATE to beginning of jump table
	movwf	STATE			;16
	goto	$+1			;17,18
	nop				;19
	retlw	0			;20,21

;************************************************************************************************************
; end21 end of second byte there must be zero
; end22 first half bit. end22 second half bit
end21:	
	btfsc	TEST_BITS,bitrec	;11
	goto	frameerr		;12
	incf	STATE,F			;13
	incf	ENDVAL,F		;14
	goto	$+1			;15,16
	goto	$+1			;17,18
	nop				;19
	retlw	0			;20,21

end22:	
	btfsc	TEST_BITS,bitrec;11
	goto	frameerr		;12
	movf	DATA5,W			;13
	movwf	DATA2			;14
	movlw	0x03			;15	point STATE to beginning of jump table
	movwf	STATE			;16
	goto	$+1			;17,18
	nop				;19
	retlw	0			;20,21

;************************************************************************************************************
; end31 end of third byte there must be a one
; end32 first half bit. end32 second half bit

end31:	
	incf	STATE,F			;11
	incf	ENDVAL,F		;12
	goto	$+1			;13,14
	goto	$+1			;15,16
	goto	$+1			;17,18
	nop				;19
	retlw	0			;20,21

end32:	
	movf	DATA5,W			;11
	movwf	DATA3			;12
	btfsc	TEST_BITS,bitrec;13	if 0 other bytes will follow
	goto	end32x			;14,15
	movlw	0x03			;15	point STATE to beginning of jump table
	movwf	STATE			;16
	goto	$+1			;17,18
	nop				;19
	retlw	0			;20,21

end32x: 
	clrf	STATE			;16	reset STATE for next preamble
	clrf	ENDVAL			;17
	clrf	DATA4			;18
	clrf	DATA5			;19
	bcf	TEST_BITS,byte4
	goto	decode			;21,22	no need here for precise cycle counting, because we decode now 
;************************************************************************************************************
; end41 end of third byte there must be a one
; end42 first half bit. end42 second half bit

end41:	
	incf	STATE,F			;11
	incf	ENDVAL,F		;12
	goto	$+1			;13,14
	goto	$+1			;15,16
	goto	$+1			;17,18
	nop				;19
	retlw	0			;20,21

end42:	
	movf	DATA5,W			;11
	movwf	DATA4			;12
	btfsc	TEST_BITS,bitrec	;13	if 0 other bytes will follow
	goto	end42x			;14,15
	movlw	0x03			;15	point STATE to beginning of jump table
	movwf	STATE			;16
	goto	$+1			;17,18
	nop				;19
	retlw	0			;20,21

end42x: 
	bsf	TEST_BITS,byte4
	clrf	STATE			;16	reset STATE for next preamble
	clrf	ENDVAL			;17
	clrf	DATA5			;18
	goto	decode			;20,21	no need to count, because we decode now 
;************************************************************************************************************
; end51 end of third byte there must be a one
; end52 first half bit. end52 second half bit

end51:	
	btfss	TEST_BITS,bitrec	;11
	goto	frameerr		;12,13
	incf	STATE,F			;13
	incf	ENDVAL,F		;14
	goto	$+1			;15,16
	goto	$+1			;17,18
	nop				;19
	retlw	0			;20,21

end52:
	btfss	TEST_BITS,bitrec	;11	test correct ending
	goto	frameerr		;12,13
	clrf	STATE			;13	reset STATE for next preamble
	clrf	ENDVAL			;14
	bsf	TEST_BITS,byte4
	goto	decode			;

;*************DECODING*******************************
decode:
; get and store buttons state
	#if mypic == 627
		clrf	BTN_STATE_PRE		; get buttons state
		btfsc	AddressButton		; bouton au +
		bsf		AddrBtnPreFlag
		btfsc	TimeButton			; bouton au +
		bsf		TimeBtnPreFlag
	#else
		clrf	BTN_STATE_PRE		; get buttons state
		btfss	AddressButton		; bouton au +
		bsf		AddrBtnPreFlag
		btfss	TimeButton			; bouton au +
		bsf		TimeBtnPreFlag
	#endif
; verify packet CRC
	movf	DATA1,W			; Exclusive or check
	xorwf	DATA2,W
	xorwf	DATA3,W
	xorwf	DATA4,W
	xorwf	DATA5,W
	btfss	STATUS,Z
	goto	exit_decode		; error in packet
; packet correct	
	call	decode_basic_acc_addr	; compute switch address
; retour avec dans W l'adresse envoye.
;	movwf	PORTB	; affichage 6 LSBs sur LEDs.
	;ici l'affichage ne reste pas parce qu'on passe ici aussi avec W=0
	iorlw	0					; check if address is valid (<>0):
	btfsc	STATUS,Z			; skip if not zero
	goto	exit_decode			; exit if zero
	movf	BTN_STATE_PRE,F		; were any buttons pressed?
	btfsc	STATUS,Z
	goto	normal_operation
	movwf	TEMP
	clrf	BTN_STATE_POST		; get buttons state again
	#if mypic == 627
		btfsc	AddressButton		; bouton au +
		bsf		AddrBtnPostFlag
		btfsc	TimeButton			; bouton au +
		bsf		TimeBtnPostFlag
	#else
		btfss	AddressButton		; bouton au +
		bsf		AddrBtnPostFlag
		btfss	TimeButton			; bouton au +
		bsf		TimeBtnPostFlag
	#endif
	
	movf	BTN_STATE_PRE,W		; check if buttons state is the same
	xorwf	BTN_STATE_POST,W
	btfsc	STATUS,Z
	goto	program_acc			; yes: jump to programming

	movf	TEMP,W				; no: continue
normal_operation:
	xorwf	ACC_ADDR,W	; test if address is ours
	btfss	STATUS,Z
	goto	exit_decode
; here address is valid and ours
	movf	DATA2,W		; extract output pair ( DATA2 = 1AAACDDD)
	andlw	1		; W = D0 = output pair
	call	do_switch
	goto	exit_decode

program_acc:							; PROGRAMMING
	clrf	ACTIVE_PAIR					; reset active pair
	movf	TEMP,W
	btfsc	AddrBtnPostFlag				;on utilise les flags
	goto	program_acc_addr
	btfss	TimeBtnPostFlag				;on utilise les flags
	goto	exit_decode

; program switch on-time or position A
	#if mypic == 627
		btfsc	AuxButton					; bouton au +
	#else
		btfss	AuxButton					; bouton au +
	#endif
	goto	ProgramPositionA
	; program switch on-time
	movwf	ACC_ON_TIME			; W -> memory
	store_W_in_eeprom_at acc_time_ee
	goto	exit_decode

ProgramPositionA:
	movwf	ServoPW_A
	store_W_in_eeprom_at ServoPW_A_ee
	; now do it
	call activate_out_A	;
	goto	exit_decode

program_acc_addr:	;program address or position B
	#if mypic == 627
		btfsc	AuxButton					; bouton au +
	#else
		btfss	AuxButton					; bouton au +
	#endif
	goto	ProgramPositionB
	; program address
	movwf	ACC_ADDR
	call	store_acc_addr
	goto	exit_decode

ProgramPositionB
	movwf	ServoPW_B
	store_W_in_eeprom_at ServoPW_B_ee
	call activate_out_A	;
;	goto	exit_decode

exit_decode:
	clrf	TEMP
	clrf	DATA1
	clrf	DATA2
	clrf	DATA3
	clrf	DATA4
	clrf	DATA5
	clrf	ENDVAL
	bcf	TEST_BITS,byte4
	return

decode_basic_acc_addr:	; returns decoded accessory address in W
; 0 means invalid. We support addresses up to (64*4)-1 = 255
; vu du LH100: 1  255 sauf 252!
; verify that DATA1 = 10AAAAAA
	btfss	DATA1,7		; bit 7 must be 1
	retlw	0
	btfsc	DATA1,6		; bit 6 must be 0
	retlw	0
; verify that DATA2 = 1AAACDDD
	btfss	DATA2,7		; bit 7 must be 1
	retlw	0
; check if address is in our supported range
; ie check that MSBs AAA = 111 (= 000 inverted)
	swapf	DATA2,W		; W = CDDD1AAA		
	andlw	b'00001111'	; W = 00001AAA
	sublw	b'00001111'	; W = W - 00001111
	btfss	STATUS,Z	; skip if 1AAA = 1111
	retlw	0			; exit if 1AAA <> 1111
; make 8-bit address.  y = carry
	rlf		DATA1,W		; DATA1 = W = 0AAAAAAy
	movwf	TEMP		; TEMP = W = 0AAAAAAy
	rlf		TEMP,W		; TEMP = W = AAAAAAyy
	andlw	b'11111100'	; W = AAAAAA00
	movwf	TEMP		; TEMP = W = AAAAAA00
	rrf		DATA2,W		; W = yCDDD1AA (A0 discarded)
	andlw	b'00000011'	; W = 000000AA
	iorwf	TEMP,W		; W = 000000AA OR AAAAAA00 = AAAAAAAA
	addlw	1			; W = W + 1
;ici l'affichage reste: affichage de quelques valeurs:
; affichage 6 LSBs de l'adresse sur LEDs
	;movwf	PORTB
; affichage CDDD: les valeurs utiles pour un dcodeur  8 sorties		
;	movwf	TEMP		; sauvegarde de W
;	movf	DATA2,W		; W = DATA2 = 1AAACDDD
;	andlw 	b'00001111'	; W = 0000CDDD
;	movwf	PORTB		; affichage
;	movf	TEMP, W		; rcupration W
	return

;*****************************************************************************
check_power_up:

	movlw	acc_time_ee		; Load
	call	read_eeprom		; switch on-time
	movwf	ACC_ON_TIME

	movlw	ServoPW_A_ee	; Load
	call	read_eeprom		; position A
	movwf	ServoPW_A

	movlw	ServoPW_B_ee	; Load
	call	read_eeprom		; position B
	movwf	ServoPW_B

	movlw	acc_addr_ee		; Load
	call	read_eeprom		; accessory address
	movwf	ACC_ADDR
	incfsz	ACC_ADDR,F		; address stored is actual acc_addr minus  one
	goto	exit_power_up	; 0x0 means first time run (invalid address)
;valeurs pour un programme vierge (eeprom=0xFF)
	movlw	1				; Default accessory address
	movwf	ACC_ADDR	
	call	store_acc_addr	; fait -1 avant d'crire en eeprom

	movlw	D'252'			; default switch on time (was 1)
	movwf	ACC_ON_TIME
	store_W_in_eeprom_at acc_time_ee

	movlw	D'32'			; default position A
	movwf	ServoPW_A
	store_W_in_eeprom_at ServoPW_A_ee

	movlw	D'224'			; default position B
	movwf	ServoPW_B
	store_W_in_eeprom_at ServoPW_B_ee

exit_power_up:
	return

store_acc_addr:		; accessory address in ACC_ADDR
	decf	ACC_ADDR,W
	store_W_in_eeprom_at acc_addr_ee
	return

read_eeprom:		; Address in W, read byte in W
	bsf	STATUS,RP0		; Bank 1
	movwf	EEADR
	bsf	EECON1,RD
	movf	EEDATA,W
	bcf	STATUS,RP0		; Bank 0
	return

write_eeprom:	; Address in W, data byte in WR_EEDATA
	bsf	STATUS,RP0		; Bank 1
	movwf	EEADR
	bcf	STATUS,RP0		; Bank 0
	movf	WR_EEDATA,W
	bsf	STATUS,RP0		; Bank 1
	movwf	EEDATA
	; begin EEPROM write required sequence
	bsf	EECON1,WREN		; Enable EEPROM write
	bcf	INTCON,GIE		; Disable interrupts
	movlw	0x55
	movwf	EECON2
	movlw	0xAA
	movwf	EECON2
	bsf	EECON1,WR
	; end sequence
	btfsc	EECON1,WR		; wait till write completes
	goto	$-1
	bsf	INTCON,GIE		; Enable interrupts
	bcf	STATUS,RP0		; Bank 0
	return

;**********************************************************************
do_switch:					; W is pair id (= D0 = coil )
; on reste dans cette routine tout le temps d'activation du servo
; aussi on peut beaucoup simplifier:
; here we test if the output must be (re)activated 
; (active_pair,7) = 0 =  output not active or retriggerable 
; (active_pair,7) = 1 = output active and not retriggerable 
	btfsc	ACTIVE_PAIR,7	; test if (flag) active pair is valid (set)
	goto	test_active_pair; = 1: do not re-activate, except if command is for the other coil 
	btfsc	ACC_ON_TIME,0	; = 0: before activating, check if is output re-triggerable
	bsf		ACTIVE_PAIR,7	; No: this is a first activation: disable re-trigger. (Yes: this is a re-triggering) 
	goto	activate_pair	; and activate
test_active_pair:			; the same coil should not be re-activated, the other yes 
	movwf	TEMP			; save W
	xorwf	ACTIVE_PAIR,W	; compare D0 (= coil, in W) to active_pair 
	andlw	b'01111111'		; (but bit 7) 
	btfsc	STATUS,Z		; not equal...
	return					; ...asked to activate already active coil: return
	movf	TEMP,W			; if not, restore W and activate
activate_pair:
	iorlw	0				; test W ( = D0 = coil )
	btfsc	STATUS,Z		; if W=0, activate A
	goto	activate_out_B	; else activate B
activate_out_A:
	#if mypic == 627
		bcf		LED6
		bsf		LED5
	#endif
	bsf		LightOut
	movf	ServoPW_A,W	;
	movwf	ServoPWM		;
	bsf		ACTIVE_PAIR,0	; remember A is active
	goto	set_up_timer
activate_out_B:
	#if mypic == 627
		bcf		LED5
		bsf		LED6
	#endif
	bcf		LightOut
	movf	ServoPW_B,W	;
	movwf	ServoPWM		;
	bcf		ACTIVE_PAIR,0	; remember B is active
set_up_timer:
	clrf	ON_TIMER_LO		; clear timer lo 
	movf	ACC_ON_TIME,W
	movwf	ON_TIMER_HI		; load timer hi 
	movlw	D'3'
	movwf	Scaler
;	clrf	ON_TIMER_CTRL	; disable timer
;	incf	ON_TIMER_HI,W	; test if on-time=255.
;	btfss	STATUS,Z		; (255+1=0) means constant output: do not enable timer 
;;	bsf		ON_TIMER_CTRL,7	; enable timer (inutile puisqu'on le reset ensuite...)
;---------------------------- gnration du signal servo
; Servo X: 0,5 ms  2,3ms = 1800 us -> 7us par pas (255 pas)
; Servo Conrad: 0,6/1,6  (mesur: 0,6/1,9 pour +/- 90)
; point milieu = 1,5ms = 1500 instructions = 8 + 128 x boucle => boucle = 11,6us
CommandeServo:
	bsf		ServoOut
	movlw	D'160'
	movwf	ZeroCtr
	decfsz	ZeroCtr,F
	goto	$-1
WaitOn:					
	decfsz	ServoPWM,F			; decrement pulse width...
	goto	$+2					; pas zero: pas de reset sortie
	bcf		ServoOut		; zero: reset sortie
	decfsz	ON_TIMER_LO,F		; ... and decrement timer low (both count 255)
	goto	WaitOn				; timer low not zero
; timer low arrive  zro: le scaler muliplie la tempo 
	decfsz	Scaler,F				; decrement scaler
	goto	WaitOn				; not zero: loop
	decfsz	ON_TIMER_HI,F		; zero: decrement timer hi... and reload scaler
	goto	ReloadScaler		; timer hi not zero: loop
;	clrf	PORTB				; timer hi at zero: reset general LEDs et servo
	#if mypic == 627
		bcf		LED6
		bcf		LED5
	#endif
	bcf	ServoOut
	return						; and exit

ReloadScaler:
	movlw	D'3'
	movwf	Scaler
	goto	CommandeServo

; fin de timer
;	bcf		ServoOut		;attention: extra pulse <<<
;	bcf		ON_TIMER_CTRL,7
;verif contenu ON_TIMER_CTRL.7 <<<
;	bcf		PORTB,0		;extinction
;	btfsc	ON_TIMER_CTRL,7
;	bsf		PORTB,0		;allumage
;	return
;----------------------------------------------------------------------------
	end

